/*
 * Copyright (c) 2010 Douglas Gilbert.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/************************************************************************
 * gpio_sysfs.c
 * 
 * Utility for testing AT91SAM9G20 (G20) and AT91SAM9G45 (G45) GPIO pins.
 * Can scan, read, set and toggle GPIO lines. The G20 has 3 banks of 32
 * pins: PA0-PA31, PB0-PB31 and PC0-PC31 while the G45 has 2 additional
 * banks: PD0-31 and PE0-31 . Some GPIO pins may be committed to other
 * uses (e.g. on the FoxG20 PB4 is TXD0 (part of a serial port)).
 * Uses sysfs gpio(lib) interface.
 *
 ****************************************************/

#define _XOPEN_SOURCE 500

#include <stdio.h>     
#include <stdlib.h>     
#include <ctype.h>     
#include <string.h>     
#include <unistd.h>    
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>


static const char * version_str = "1.05 20100218";

#define EXPORT_FILE "/sys/class/gpio/export"
#define UNEXPORT_FILE "/sys/class/gpio/unexport"
#define GPIO_BASE_FILE "/sys/class/gpio/gpio"

#define DEF_NUM_TOGGLE 1000000
#define GPIO_BANKS_G20 3    /* PA0-31,PB0-31 and PC0-31 */
#define GPIO_BANKS_G45 5    /* as for G20 plus PD0-31 and PE0-31 */
#define LINES_PER_BANK 32


static struct timespec delay_req = { .tv_sec = 0, .tv_nsec = 0 };

static int verbose = 0;


static void
usage(void)
{
    fprintf(stderr, "Usage: "
            "gpio_sysfs [-b <bn>] [-d <usec>] [-e] [-f] [-h] [-n <num>]\n"
            "                  [-p <port>] [-r] [-s 0|1] [-S] [-t] [-u] "
            "[-v] [-V]\n"
            "  where:\n"
            "    -b <bn>      bit number within port (0 to 31)\n"
            "    -d <usec>    delay between toggling gpio levels (def: 0)\n"
            "    -e           enumerate pin names with corresponding "
            "kernel pin\n"
            "                 numbers (once for G20, twice for G45)\n"
            "    -f           force (drive) high when toggling (default "
            "for high\n"
            "                 is input mode and assume pullup)\n"
            "    -h           print usage message\n"
            "    -n <num>     number of cycles to toggle gpio line (def: "
            "1000000)\n"
            "    -p <port>    port bank ('A' to 'E') or gpio kernel line "
            "number\n"
            "    -r           read gpio line and send value to stdout\n"
            "                 used twice: exit value 0 for low, 1 for "
            "high\n"
            "    -s 0|1       set GPIO line to 0 (low) or 1 (high)\n"
            "    -S           scan all pins for availability\n"
            "                 use once for G20, twice for G45\n"
            "    -t           toggle gpio line <num> times\n"
            "    -u           unexport gpio line\n"
            "    -v           increase verbosity (multiple times for more)\n"
            "    -V           print version string then exit\n\n"
            "AT91SAM9G20 (and G45) GPIO test program. Uses sysfs interface.\n"
            "Can scan ('-S'), set and read lines. Can toggle line ('-t') "
            "<num> times\nwith <usec> delay prior to each transition.\n");
}

static int
gs_export(int * exp_fdp, int knum)
{
    int res;
    char b[64];

    snprintf(b, sizeof(b), "%d", knum);
    res = pwrite(*exp_fdp, b, strlen(b), 0);
    if (res < 0)
        fprintf(stderr, "Unable to export knum=%d (already in use?): %s\n",
                knum, strerror(errno));
    return res;
}

static int
gs_unexport(int * unexp_fdp, int knum)
{
    int res;
    char b[64];

    snprintf(b, sizeof(b), "%d", knum);
    res = pwrite(*unexp_fdp, b, strlen(b), 0);
    if (res < 0)
        fprintf(stderr, "Unable to unexport knum=%d: %s\n", knum,
                strerror(errno));
    return res;
}


int
main(int argc, char ** argv)
{
    int opt, k, j, n, num, res, exp_fd, unexp_fd, dir_fd, val_fd, ok, avail;
    int delay = 0;
    int enumerate = 0;
    int force = 0;
    int knum = -1;
    int num_toggle = DEF_NUM_TOGGLE;
    int read_val = 0;
    int scan = 0;
    int state = -1;
    int toggle = 0;
    int unexport = 0;
    int ret = -1;
    int bit_num = -1;
    char b[64];
    char ch;
    char bank = '\0';

    while ((opt = getopt(argc, argv, "b:d:efhn:p:rs:StuvV")) != -1) {
        switch (opt) {
        case 'b':
            k = atoi(optarg);
            if ((k < 0) || (k > 31)) {
                fprintf(stderr, "'-b' expects a bit number from 0 to 31\n");
                exit(EXIT_FAILURE);
            }
            bit_num = k;
            break;
        case 'd':
            k = atoi(optarg);
            if (k < 0) {
                fprintf(stderr, "'-d' expects a delay in microseconds from "
                        "0 to 2147483647\n");
                exit(EXIT_FAILURE);
            }
            delay_req.tv_sec = k / 1000000;     /* seconds component */
            delay_req.tv_nsec = (k % 1000000) * 1000;  /* nanoseconds */
            delay = (delay_req.tv_sec || delay_req.tv_nsec);
            break;
        case 'e':
            ++enumerate;
            break;
        case 'f':
            ++force;
            break;
        case 'h':
            usage();
            exit(EXIT_SUCCESS);
            break;
        case 'n':
            num_toggle = atoi(optarg);
            break;
        case 'p':
            if (isalpha(*optarg)) {
                ch = toupper(*optarg);
                if ((ch >= 'A') && (ch <= 'E'))
                    bank = ch;
                else {
                    fprintf(stderr, "'-p' expects a letter ('A' to 'E')\n");
                    exit(EXIT_FAILURE);
                }
            } else if (isdigit(*optarg)) {
                k = atoi(optarg);
                if ((k < 32) || (k > 511)) {
                    fprintf(stderr, "'-p' expects a letter or a number "
                            "32 or greater\n");
                    exit(EXIT_FAILURE);
                }
                knum = k;
            } else {
                fprintf(stderr, "'-p' expects a letter ('A' to 'E') or "
                        "a number\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 'r':
            ++read_val;
            break;
        case 's':
            state = atoi(optarg);
            if ((state < 0) || (state > 1)) {
                fprintf(stderr, "'-s' expects '0' or '1'\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 'S':
            ++scan;
            break;
        case 't':
            ++toggle;
            break;
        case 'u':
            ++unexport;
            break;
        case 'v':
            ++verbose;
            break;
        case 'V':
            printf("%s\n", version_str);
            exit(EXIT_SUCCESS);
        default: /* '?' */
            usage();
            exit(EXIT_FAILURE);
        }
    }
    if (optind < argc) {
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            exit(EXIT_FAILURE);
        }
    }

    if (enumerate) {
        num = (1 == enumerate) ? GPIO_BANKS_G20 : GPIO_BANKS_G45;
        for (k = 0; k < LINES_PER_BANK; ++k) {
            for (j = 0; j < num; ++j) {
                n = ((j + 1) * 32) + k;
                printf("%sP%c%d: %d   ", (j ? "\t" : ""), 'A' + j, k, n);
            }
            printf("\n");
        }
        return 0;
    }

    if ((knum >= 0) || (scan)) {
        if (bit_num >= 0) {
            fprintf(stderr, "Give either '-p <knum>' or ('-b <bn>' and "
                    "'-p <bank>') but not both\n");
            exit(EXIT_FAILURE);
        }
    } else if (bank) {
        if (bit_num < 0) {
            fprintf(stderr, "If '-p <bank>' given then also need "
                    "'-b <bn>'\n");
            exit(EXIT_FAILURE);
        } else
            knum = ((bank + 1 - 'A') * 32) + bit_num;
    } else {
        fprintf(stderr, "Need to give gpio line with '-p <port>' and/or "
                "'-b <bn>'\n");
        usage();
        exit(EXIT_FAILURE);
    }

    if (! (scan || toggle || unexport || read_val || (state >= 0))) {
        fprintf(stderr, "Expect at least '-r', '-s', '-S', '-t' or '-u' "
                "option, use '-h' for usage\n");
        usage();
        exit(EXIT_FAILURE);
    }
    if (read_val && (toggle || scan || (state >= 0))) {
        fprintf(stderr, "Can't have '-r' with '-s', -S' or  '-t'\n");
        usage();
        exit(EXIT_FAILURE);
    }

    exp_fd = -1;
    unexp_fd = -1;
    dir_fd = -1;
    val_fd = -1;
    exp_fd = open(EXPORT_FILE, O_WRONLY);
    if (exp_fd < 0) {
        perror(EXPORT_FILE);
        goto bad;
    }
    unexp_fd = open(UNEXPORT_FILE, O_WRONLY);
    if (unexp_fd < 0) {
        perror(UNEXPORT_FILE);
        goto bad;
    }

    if (unexport) {
        if (gs_unexport(&unexp_fd, knum) < 0)
            fprintf(stderr, "continue ...\n");
    }

    if (scan) {
        num = (1 == scan) ? GPIO_BANKS_G20 : GPIO_BANKS_G45;
        for (k = 0, avail = 0; k < LINES_PER_BANK; ++k) {
            for (j = 0; j < num; ++j) {
                n = ((j + 1) * 32) + k;
                snprintf(b, sizeof(b), "%d", n);
                ok = (pwrite(exp_fd, b, strlen(b), 0) >= 0);
                printf("%sP%c%d %s", (j ? "\t" : ""), 'A' + j, k,
                       (ok ? "AVAIL" : "occup"));
                if (ok) {
                    ++avail;
                    gs_unexport(&unexp_fd, n);
                }
            }
            printf("\n");
        }
        printf("%d GPIO lines are available via sysfs\n", avail);
    }

    if (toggle) {
        if (verbose)
            fprintf(stderr, "Toggling knum=%d\n", knum);
        if (gs_export(&exp_fd, knum) < 0)
            goto bad;
        snprintf(b, sizeof(b), "%s%d/direction", GPIO_BASE_FILE, knum);
        dir_fd = open(b, O_WRONLY);
        if (dir_fd < 0) {
            fprintf(stderr, "Open %s: %s\n", b, strerror(errno));
            goto bad;
        }
        snprintf(b, sizeof(b), "%s%d/value", GPIO_BASE_FILE, knum);
        val_fd = open(b, O_WRONLY);
        if (val_fd < 0) {
            fprintf(stderr, "Open %s: %s\n", b, strerror(errno));
            goto bad;
        }
        if (pwrite(dir_fd, "out\n", 4, 0) < 0) {
            fprintf(stderr, "Unable to write out to dir_fd: %s\n",
                    strerror(errno));
            goto bad;
        }
        if (delay && verbose)
            fprintf(stderr, "After each edge delay for %d.%06ld seconds\n",
                    (int)delay_req.tv_sec, delay_req.tv_nsec / 1000);

        for (k = 0; k < num_toggle; ++k) {
            if (force) {
                if (pwrite(val_fd, "0", 1, 0) < 0) {
                    fprintf(stderr, "Unable to pwrite 0 to val_fd: %s\n",
                            strerror(errno));
                    goto bad;
                }
                if (delay)
                    nanosleep(&delay_req, NULL);
                if (pwrite(val_fd, "1", 1, 0) < 0) {
                    fprintf(stderr, "Unable to pwrite 1 to val_fd: %s\n",
                            strerror(errno));
                    goto bad;
                }
                if (delay)
                    nanosleep(&delay_req, NULL);
            } else {
                if (pwrite(dir_fd, "low", 3, 0) < 0) {
                    fprintf(stderr, "Unable to pwrite low to dir_fd: %s\n",
                            strerror(errno));
                    goto bad;
                }
                if (delay)
                    nanosleep(&delay_req, NULL);
                if (pwrite(dir_fd, "in", 2, 0) < 0) {
                    fprintf(stderr, "Unable to pwrite in to dir_fd: %s\n",
                            strerror(errno));
                    goto bad;
                }
                if (delay)
                    nanosleep(&delay_req, NULL);
            }
        }
        if (state >= 0) {
            if (0 == state)
                res = pwrite(dir_fd, "low", 3, 0);
            else
                res = pwrite(dir_fd, "high", 4, 0);
            if (res < 0) {
                fprintf(stderr, "Unable to pwrite low/high to dir_fd: %s\n",
                        strerror(errno));
                goto bad;
            }
        }
        if (gs_unexport(&unexp_fd, knum) < 0)
            goto bad;
        ret = 0;
    } else if (state >= 0) {
        if (gs_export(&exp_fd, knum) < 0)
            goto bad;
        snprintf(b, sizeof(b), "%s%d/direction", GPIO_BASE_FILE, knum);
        dir_fd = open(b, O_WRONLY);
        if (dir_fd < 0) {
            fprintf(stderr, "Open %s: %s\n", b, strerror(errno));
            goto bad;
        }
        if (0 == state)
            res = pwrite(dir_fd, "low", 3, 0);
        else
            res = pwrite(dir_fd, "high", 4, 0);
        if (res < 0) {
            fprintf(stderr, "Unable to pwrite low/high to dir_fd: %s\n",
                    strerror(errno));
            goto bad;
        }
        if (gs_unexport(&unexp_fd, knum) < 0)
            goto bad;
        ret = 0;
    } else if (read_val) {
        if (gs_export(&exp_fd, knum) < 0)
            goto bad;
        snprintf(b, sizeof(b), "%s%d/direction", GPIO_BASE_FILE, knum);
        dir_fd = open(b, O_WRONLY);
        if (dir_fd < 0) {
            fprintf(stderr, "Open %s: %s\n", b, strerror(errno));
            goto bad;
        }
        snprintf(b, sizeof(b), "%s%d/value", GPIO_BASE_FILE, knum);
        val_fd = open(b, O_RDONLY);
        if (val_fd < 0) {
            fprintf(stderr, "Open %s: %s\n", b, strerror(errno));
            goto bad;
        }
        if (pwrite(dir_fd, "in", 2, 0) < 0) {
            fprintf(stderr, "Unable to pwrite in to dir_fd: %s\n",
                    strerror(errno));
            goto bad;
        }
        k = pread(val_fd, b, 1, 0);
        if (k < 0) {
            fprintf(stderr, "pread() failed: %s\n", strerror(errno));
            goto bad;
        } else if (0 == k) {
            fprintf(stderr, "pread could not find a value\n");
            goto bad;
        }
        printf("%c\n", b[0]);
        if (read_val > 1)
            ret = ('0' !=  b[0]);
        else
            ret = 0;
        if (gs_unexport(&unexp_fd, knum) < 0) {
            ret = 1;
            goto bad;
        }
    } else
        ret = 0;

bad:
    if (val_fd >= 0)
        close(val_fd);
    if (dir_fd >= 0)
        close(dir_fd);
    if (unexp_fd >= 0)
        close(unexp_fd);
    if (exp_fd >= 0)
        close(exp_fd);
    if (ret)
        exit(EXIT_FAILURE);
    else
        return 0;
}
